; File: 	dcf77.asm
; Device:	AVR
; Created:	2023-10-31
; Version:	2023-11-02
; Author:	Johannes Fechner

.equ DCF_STATUS_FILTERED = 4	; Bit position of filtered DCF signal inside dcf_status register.
.equ DCF_STATUS_IN_SYNC = 5		; Bit position of in-synchronization flag inside dcf_status register.
.equ DCF_STATUS_EVENT = 6		; Bit position of event flag inside dcf_status register.
.equ DCF_STATUS_0BIT = $0		; Event value of "'0' bit received" inside low nibble of dcf_status register.
.equ DCF_STATUS_1BIT = $1		; Event value of "'1' bit received" inside low nibble of dcf_status register.
.equ DCF_STATUS_PAUSE = $2		; Event value of "sync pause received" inside low nibble of dcf_status register.
.equ DCF_STATUS_ERROR = $3		; Event value of "error occurred" inside low nibble of dcf_status register.

.if DCF_UART_DEBUG
; === UART debug message codes ===
.equ DCF_PULSE_TOO_SHORT = 0
.equ DCF_PULSE_BETWEEN_0_1 = 1
.equ DCF_PULSE_TOO_LONG = 2
.equ DCF_ERROR_CODE_COUNT = 3
.endif

; === Calculations for DCF signal filtering ===
; Set the filter sampling distance in seconds:
#define DCF_FILTER_SET_DIST 1.0e-3
.equ DCF_FILTER_PRESC = 256 ; Prescaling divisor.
.equ DCF_FILTER_OCR = ROUND(DCF_FILTER_SET_DIST*F_CPU/DCF_FILTER_PRESC)-1 ; == 28.
; Calculate the real resulting interrupt distance in seconds:
#define DCF_FILTER_REAL_DIST (1.0*DCF_FILTER_PRESC*(DCF_FILTER_OCR+1)/F_CPU)
; Set the acceptable pulse time tolerance in percent:
.equ DCF_PULSE_TOL = 20
; Calculate the approximate interrupt counts of time constants (T is time in seconds):
#define DCF_MIN_INTR(T) ROUND((100.0-DCF_PULSE_TOL)*(T)/DCF_FILTER_REAL_DIST/100.0)
#define DCF_MAX_INTR(T) ROUND((100.0+DCF_PULSE_TOL)*(T)/DCF_FILTER_REAL_DIST/100.0)+1
#define DCF_INTR(T) ROUND(1.0*(T)/DCF_FILTER_REAL_DIST)
; Symbols for the calculated interrupt counts:
.equ DCF_PULSE_0_MIN_INTR = DCF_MIN_INTR(0.1)
.equ DCF_PULSE_0_MAX_INTR = DCF_MAX_INTR(0.1)
.equ DCF_PULSE_1_MIN_INTR = DCF_MIN_INTR(0.2)
.equ DCF_PULSE_1_MAX_INTR = DCF_MAX_INTR(0.2)
.equ DCF_PAUSE = DCF_INTR(1.5)

; === Routine: DCF decoder initialization ===
dcf_init:
	clr		dcf_filter
	clr		dcf_status
	clr		dcf_isrCntrL
	clr		dcf_isrCntrH
.if DCF_TEST_OUTPUT
	sbi		DCF_TEST_OUTPUT_DDR, DCF_TEST_OUTPUT_POS
.endif ; DCF_TEST_OUTPUT
	ret

; === Interrupt routine: DCF decoder ===
dcf_isr:
	push	temp0
	push	temp1
	push	temp2
	xin		temp0, SREG
	push	temp0
; Increment interrupt counter, preventing overflow:
	movw	temp1:temp0, dcf_isrCntrH:dcf_isrCntrL
	subi	temp0, $FF
	sbci	temp1, $FF
	breq	dcf_isr_noInc	; Maximum counter value reached.
	movw	temp1:temp0, dcf_isrCntrH:dcf_isrCntrL
	subi	temp0, low(-1)
	sbci	temp1, high(-1)
	movw	dcf_isrCntrH:dcf_isrCntrL, temp1:temp0
dcf_isr_noInc:
; Read Analog Comparator output:
	xin		temp0, ACSR
	sbrc	temp0, ACO
	rjmp	dcf_isr_aco1
; AC output is 0.
	tst		dcf_filter
	breq	dcf_isr_bottom	; Filter value is already 0.
	dec		dcf_filter			; Else, decrement filter value.
	rjmp	dcf_isr_end
dcf_isr_bottom:
	sbrs	dcf_status, DCF_STATUS_FILTERED
	rjmp	dcf_isr_end	; Filter output was already 0.
; Negative edge detected.
	clt							; Update filter output:
	bld		dcf_status, DCF_STATUS_FILTERED
	inc		dcf_bitCntr
.if DCF_TEST_OUTPUT
	cbi		DCF_TEST_OUTPUT_PORT, DCF_TEST_OUTPUT_POS
.endif ; DCF_TEST_OUTPUT
; Test whether a synchronization pause was received:
	movw	temp1:temp0, dcf_isrCntrH:dcf_isrCntrL
	subi	temp0, low(DCF_PAUSE)
	sbci	temp1, high(DCF_PAUSE)
	brlo	dcf_isr_clrIsrCntr	; No pause.
; A pause was received.
	set
	bld		dcf_status, DCF_STATUS_IN_SYNC
	bld		dcf_status, DCF_STATUS_EVENT
	mov		temp0, dcf_status
	andi	temp0, $F0
	ori		temp0, DCF_STATUS_PAUSE
	mov		dcf_status, temp0
	clr		dcf_bitCntr
	rjmp	dcf_isr_clrIsrCntr
dcf_isr_aco1:
; AC output is 1.
	mov		temp0, dcf_filter
	cpi		temp0, DCF_FILTER_MAX_VAL
	breq	dcf_isr_top	; Filter value is already at maximum.
	inc		dcf_filter			; Else, increment filter value.
	rjmp	dcf_isr_end
dcf_isr_top:
	sbrc	dcf_status, DCF_STATUS_FILTERED
	rjmp	dcf_isr_end	; Filter output was already 1.
; Positive edge detected.
	set							; Update filter output:
	bld		dcf_status, DCF_STATUS_FILTERED
.if DCF_TEST_OUTPUT
	sbi		DCF_TEST_OUTPUT_PORT, DCF_TEST_OUTPUT_POS
.endif ; DCF_TEST_OUTPUT
; Evaluate interrupt counter:
	movw	temp1:temp0, dcf_isrCntrH:dcf_isrCntrL
.if DCF_UART_DEBUG
	ldi		temp2, DCF_PULSE_TOO_SHORT
.endif ; DCF_UART_DEBUG
	subi	temp0, low(DCF_PULSE_0_MIN_INTR)
	sbci	temp1, high(DCF_PULSE_0_MIN_INTR)
	brlo	dcf_isr_error
	movw	temp1:temp0, dcf_isrCntrH:dcf_isrCntrL
	subi	temp0, low(DCF_PULSE_0_MAX_INTR)
	sbci	temp1, high(DCF_PULSE_0_MAX_INTR)
	brlo	dcf_isr_0bit
	movw	temp1:temp0, dcf_isrCntrH:dcf_isrCntrL
.if DCF_UART_DEBUG
	ldi		temp2, DCF_PULSE_BETWEEN_0_1
.endif ; DCF_UART_DEBUG
	subi	temp0, low(DCF_PULSE_1_MIN_INTR)
	sbci	temp1, high(DCF_PULSE_1_MIN_INTR)
	brlo	dcf_isr_error
	movw	temp1:temp0, dcf_isrCntrH:dcf_isrCntrL
	subi	temp0, low(DCF_PULSE_1_MAX_INTR)
	sbci	temp1, high(DCF_PULSE_1_MAX_INTR)
	brlo	dcf_isr_1bit
; Pulse was too long.
.if DCF_UART_DEBUG
	ldi		temp2, DCF_PULSE_TOO_LONG
.endif
dcf_isr_error:
	set
	bld		dcf_status, DCF_STATUS_EVENT
	mov		temp0, dcf_status
	andi	temp0, $F0
	ori		temp0, DCF_STATUS_ERROR
	mov		dcf_status, temp0
.if DCF_UART_DEBUG
	sts		dcf_errCode, temp2
.endif ; DCF_UART_DEBUG
	rjmp	dcf_isr_clrIsrCntr
dcf_isr_0bit:
	set
	bld		dcf_status, DCF_STATUS_EVENT
	mov		temp0, dcf_status
	andi	temp0, $F0
	ori		temp0, DCF_STATUS_0BIT
	mov		dcf_status, temp0
	rjmp	dcf_isr_clrIsrCntr
dcf_isr_1bit:
	set
	bld		dcf_status, DCF_STATUS_EVENT
	mov		temp0, dcf_status
	andi	temp0, $F0
	ori		temp0, DCF_STATUS_1BIT
	mov		dcf_status, temp0
dcf_isr_clrIsrCntr:
.if DCF_UART_DEBUG
	sts		dcf_isrCntrLCopy, dcf_isrCntrL
	sts		dcf_isrCntrHCopy, dcf_isrCntrH
.endif ; DCF_UART_DEBUG
	clr		dcf_isrCntrL
	clr		dcf_isrCntrH
dcf_isr_end:
	pop		temp0
	xout	SREG, temp0
	pop		temp2
	pop		temp1
	pop		temp0
	reti
