; File:		dcfRxTest-public-231104\main.asm
; Device:	ATmega88A
; Created:	04.11.2023 11:42:20
; Version:	2023-11-04
; Author:	Johannes Fechner


; == UART debug mode selection ==
.equ DCF_UART_DEBUG = 1
.if DCF_UART_DEBUG
; Set the UART baud rate in bits per second:
.equ UART_BPS = 9_600
.endif ; DCF_UART_DEBUG

; == DCF LED output selection ==
.equ DCF_TEST_OUTPUT = 1
.if DCF_TEST_OUTPUT
.equ DCF_TEST_OUTPUT_DDR = DDRB
.equ DCF_TEST_OUTPUT_PORT = PORTB
.equ DCF_TEST_OUTPUT_POS = 0
.endif ; DCF_TEST_OUTPUT

; == Hardware definitions ==
; === Device selection ===
.equ ATMEGAXXA_PA = 1
.equ TINY_1_SERIES = 0

; === CPU clock frequency in Hz ===
.equ F_CPU = 7_372_800

; === DCF Rx input = Analog Comparator inverting input ===
.equ ANACOMP_PIN = PINC
.equ ANACOMP = 1

; == Register definitions ==
; === Zero register ===
.def reg00 = r2

; === Temporary/general registers ===
; Must be immediate registers (r16..r25)
.def temp0 = r16
.def temp1 = r17
.def temp2 = r18
.def temp3 = r19
.def temp4 = r20
.def temp5 = r21

; === DCF receiving registers ===
.def dcf_filter = r3
.def dcf_status = r4
.def dcf_bitCntr = r5
.def dcf_isrCntrL = r6	; Must be an even register.
.def dcf_isrCntrH = r7	; Must be dcfIntrCntrL+1.

.if DCF_UART_DEBUG
; == RAM address definitions ==
.dseg
.org SRAM_START
dcf_errCode:		.byte 1
dcf_isrCntrLCopy:	.byte 1
dcf_isrCntrHCopy:	.byte 1
.cseg
.endif ; DCF_UART_DEBUG

; == Code segment ==
.org 0
    rjmp    reset
.org OC2Aaddr				; Timer2 Compare Match A
	rjmp	dcf_isr
;.org OC1Aaddr				; Timer1 Compare Match A
;	rjmp	isrSecTick
.org INT_VECTORS_SIZE

.include "macros.asm"
.if DCF_UART_DEBUG
.include "uart.asm"
.endif ; DCF_UART_DEBUG
.include "dcf77.inc"
.include "dcf77.asm"

.if DCF_UART_DEBUG
; Zero-terminated strings
szBitNo:
.db "Bit #", 0
sz0bit:
.db "Bit '0' received (", 0, 0
sz1bit:
.db "Bit '1' received (", 0, 0
szPause:
.db "Possible synchronization pause received (", 0
szError0:
.db "Error: Pulse was too short (", 0, 0
szError1:
.db "Error: Pulse was between 0 and 1 (", 0, 0
szError2:
.db "Error: Pulse was too long (", 0
szMessageEnd:
.db " ms).", '\r', '\n', 0
szInvalidErrorCode:
.db "Error: Invalid error code.", '\r', '\n', 0, 0
szInvalidEventCode:
.db "Error: Invalid event code.", '\r', '\n', 0, 0
; Define string table:
errorStringTable:
; The following entries must be left-shifted once for the LPM instruction.
.dw szError0<<1, szError1<<1, szError2<<1
.endif ; DCF_UART_DEBUG

reset:
; Initialize stack pointer:
    ldi     temp0, high(RAMEND)
    out     SPH, temp0
    ldi     temp0, low(RAMEND)
    out     SPL, temp0
; Initialize auxiliary registers:
	clr     reg00
; Initialize DCF registers:
	rcall	dcf_init

.if DCF_UART_DEBUG
; === Initialize UART: 8 data bits, no parity, 1 stop bit ===
.equ UBRR_VAL = ROUND(F_CPU/(16*UART_BPS))-1
	ldi		temp0, high(UBRR_VAL)
	xout	UBRR0H, temp0
	ldi		temp0, low(UBRR_VAL)
	xout	UBRR0L, temp0
	ldi		temp0, (1<<UCSZ01) | (1<<UCSZ00)
	xout	UCSR0C, temp0
	ldi		temp0, 1<<TXEN0							; Transmitting only.
	xout	UCSR0B, temp0
.endif ; DCF_UART_DEBUG

; === Initialize Analog Comparator ===
	ldi		temp0, ANACOMP
	xout	ADMUX, temp0
	ldi		temp0, 1<<ACME
	xout	ADCSRB, temp0
	ldi		temp0, 1<<ACBG
	xout	ACSR, temp0

; === Initialize Timer2 for DCF signal filtering ===
	ldi		temp0, DCF_FILTER_OCR
	xout	OCR2A, temp0
	ldi		temp0, 1<<WGM21 ; CTC mode.
	xout	TCCR2A, temp0
.if DCF_FILTER_PRESC == 256
	ldi		temp0, (1<<CS22) | (1<<CS21) ; Prescaling 1/256.
.else
.error "TCCR2B value must be changed according to DCF_FILTER_PRESC."
.endif
	xout	TCCR2B, temp0
	ldi		temp0, 1<<OCIE2A ; Enable Compare Match A interrupt.
	xout	TIMSK2, temp0
; Enable interrupts:
	sei
; Main loop:
loop:
	sbrs	dcf_status, DCF_STATUS_EVENT
	rjmp	loop
; DCF_STATUS_EVENT is set.
	cli
	clt
	bld		dcf_status, DCF_STATUS_EVENT
; If synchronized, print the bit number:
	bst		dcf_status, DCF_STATUS_IN_SYNC
	brtc	loop_notSynchronized
	ldiz	szBitNo<<1
	rcall	uart_putSz
	mov		temp0, dcf_bitCntr
	rcall	uart_putDecByte
	ldi		temp0, '.'
	rcall	uart_putChar
	ldi		temp0, ' '
	rcall	uart_putChar
loop_notSynchronized:
; Branch according to type of event:
	mov		temp0, dcf_status
	andi	temp0, $0F
	cpi		temp0, DCF_STATUS_0BIT
	brne	loop_1
	rjmp	loop_0bit
loop_1:
	cpi		temp0, DCF_STATUS_1BIT
	brne	loop_2
	rjmp	loop_1bit
loop_2:
	cpi		temp0, DCF_STATUS_PAUSE
	brne	loop_3
	rjmp	loop_pause
loop_3:
	cpi		temp0, DCF_STATUS_ERROR
	brne	loop_4
	rjmp	loop_error
loop_4:
; Unknown event code.
	ldiz	szInvalidEventCode<<1
	rcall	uart_putSz
	sei
	rjmp	loop
loop_error:
	lds		temp0, dcf_errCode
; Make sure that error code is valid:
	cpi		temp0, DCF_ERROR_CODE_COUNT
	brlo	loop_validErrorCode
	ldiz	szInvalidErrorCode<<1
	rcall	uart_putSz
	sei
	rjmp	loop
loop_validErrorCode:
	lsl		temp0
; Load the address of the appropriate error message string:
	ldiz	errorStringTable<<1
	addz	temp0
	lpm		temp0, Z+		; Fetch LSB.
	lpm		temp1, Z		; Fetch MSB.
	movw	ZH:ZL, temp1:temp0
	rcall	uart_putSz
; Print isr counter value in hex and time in decimal:
	ldi		temp0, '$'
	rcall	uart_putChar
	lds		temp0, dcf_isrCntrHCopy
	rcall	uart_putHexByte
	lds		temp0, dcf_isrCntrLCopy
	rcall	uart_putHexByte
	ldi		temp0, ','
	rcall	uart_putChar
	ldi		temp0, ' '
	rcall	uart_putChar
	rcall	printDelayTime
	ldiz	2*szMessageEnd
	rcall	uart_putSz
	sei
	rjmp	loop
loop_0bit:
	ldiz	2*sz0bit
	rcall	uart_putSz
; Print isr counter value in hex and time in decimal:
	ldi		temp0, '$'
	rcall	uart_putChar
	lds		temp0, dcf_isrCntrHCopy
	rcall	uart_putHexByte
	lds		temp0, dcf_isrCntrLCopy
	rcall	uart_putHexByte
	ldi		temp0, ','
	rcall	uart_putChar
	ldi		temp0, ' '
	rcall	uart_putChar
	rcall	printDelayTime
	ldiz	2*szMessageEnd
	rcall	uart_putSz
	sei
	rjmp	loop
loop_1bit:
	ldiz	2*sz1bit
	rcall	uart_putSz
; Print isr counter value in hex and time in decimal:
	ldi		temp0, '$'
	rcall	uart_putChar
	lds		temp0, dcf_isrCntrHCopy
	rcall	uart_putHexByte
	lds		temp0, dcf_isrCntrLCopy
	rcall	uart_putHexByte
	ldi		temp0, ','
	rcall	uart_putChar
	ldi		temp0, ' '
	rcall	uart_putChar
	rcall	printDelayTime
	ldiz	2*szMessageEnd
	rcall	uart_putSz
	sei
	rjmp	loop
loop_pause:
	ldiz	2*szPause
	rcall	uart_putSz
; Print isr counter value in hex and time in decimal:
	ldi		temp0, '$'
	rcall	uart_putChar
	lds		temp0, dcf_isrCntrHCopy
	rcall	uart_putHexByte
	lds		temp0, dcf_isrCntrLCopy
	rcall	uart_putHexByte
	ldi		temp0, ','
	rcall	uart_putChar
	ldi		temp0, ' '
	rcall	uart_putChar
	rcall	printDelayTime
	ldiz	2*szMessageEnd
	rcall	uart_putSz
	sei
	rjmp	loop

; === Routine: Calculate the previous DCF pulse/pause time and send it in decimal via UART ===
; Return:	temp2:temp1:temp0, time in microseconds.
.equ DCF_ISR_DIST_INT = ROUND(1.0e7*DCF_FILTER_REAL_DIST)	; Interrupt distance in 10^(-7) seconds.
printDelayTime:
	push	temp0
	push	temp1
	push	temp2
	push	temp3
	push	temp4
	push	temp5
; low * low
	ldi		temp0, low(DCF_ISR_DIST_INT)
	lds		temp1, dcf_isrCntrLCopy
	mul		temp0, temp1
	mov		temp2, r0
	mov		temp3, r1
; high * low
	ldi		temp0, high(DCF_ISR_DIST_INT)
	mul		temp0, temp1
	clr		temp4
	add		temp3, r0
	adc		temp4, r1
; high * high
	lds		temp1, dcf_isrCntrHCopy
	mul		temp0, temp1
	clr		temp5
	add		temp4, r0
	adc		temp5, r1
; low * high
	ldi		temp0, low(DCF_ISR_DIST_INT)
	mul		temp0, temp1
	add		temp3, r0
	adc		temp4, r1
; temp5:temp4:temp3:temp2 now contains the ISR distance time in 10^(-7) seconds.
; Write the number to UART in decimal ASCII:
	ldi		temp0, '0'-1
	clt							; Used for leading-zero suppression.
printDelayTime_8:
	inc		temp0
	subi	temp2, low(100_000_000)
	sbci	temp3, byte2(100_000_000)
	sbci	temp4, byte3(100_000_000)
	sbci	temp5, byte4(100_000_000)
	brcc	printDelayTime_8
	subi	temp2, low(-100_000_000)
	sbci	temp3, byte2(-100_000_000)
	sbci	temp4, byte3(-100_000_000)
	sbci	temp5, byte4(-100_000_000)
	cpi		temp0, '0'
	breq	printDelayTime_8z
	set
	rcall	uart_putChar
printDelayTime_8z:
; Value is now <= 99_999_999 = $5_F5_E0_FF.
	ldi		temp0, '0'-1
printDelayTime_7:
	inc		temp0
	subi	temp2, low(10_000_000)
	sbci	temp3, byte2(10_000_000)
	sbci	temp4, byte3(10_000_000)
	sbci	temp5, byte4(10_000_000)
	brcc	printDelayTime_7
	subi	temp2, low(-10_000_000)
	sbci	temp3, byte2(-10_000_000)
	sbci	temp4, byte3(-10_000_000)
	sbci	temp5, byte4(-10_000_000)
	brts	printDelayTime_7ts
	cpi		temp0, '0'
	breq	printDelayTime_7z
	set
printDelayTime_7ts:
	rcall	uart_putChar
printDelayTime_7z:
; Value is now <= 9_999_999 = $98_96_7F, so temp5 contains 0 for sure.
	ldi		temp0, '0'-1
printDelayTime_6:
	inc		temp0
	subi	temp2, low(1_000_000)
	sbci	temp3, byte2(1_000_000)
	sbci	temp4, byte3(1_000_000)
	brcc	printDelayTime_6
	subi	temp2, low(-1_000_000)
	sbci	temp3, byte2(-1_000_000)
	sbci	temp4, byte3(-1_000_000)
	brts	printDelayTime_6ts
	cpi		temp0, '0'
	breq	printDelayTime_6z
	set
printDelayTime_6ts:
	rcall	uart_putChar
printDelayTime_6z:
; Value is now <= 999_999 = $F_42_3F.
	ldi		temp0, '0'-1
printDelayTime_5:
	inc		temp0
	subi	temp2, low(100_000)
	sbci	temp3, byte2(100_000)
	sbci	temp4, byte3(100_000)
	brcc	printDelayTime_5
	subi	temp2, low(-100_000)
	sbci	temp3, byte2(-100_000)
	sbci	temp4, byte3(-100_000)
	brts	printDelayTime_5ts
	cpi		temp0, '0'
	breq	printDelayTime_5z
	set
printDelayTime_5ts:
	rcall	uart_putChar
printDelayTime_5z:
; Value is now <= 99_999 = $1_86_9F.
	ldi		temp0, '0'-1
printDelayTime_4:
	inc		temp0
	subi	temp2, low(10_000)
	sbci	temp3, byte2(10_000)
	sbci	temp4, byte3(10_000)
	brcc	printDelayTime_4
	subi	temp2, low(-10_000)
	sbci	temp3, byte2(-10_000)
	sbci	temp4, byte3(-10_000)
	rcall	uart_putChar
; Restore the modified registers:
	pop		temp5
	pop		temp4
	pop		temp3
	pop		temp2
	pop		temp1
	pop		temp0
	ret
